from abc import ABC
import typing_extensions

from experimaestro.core.types import ObjectType
import experimaestro
import io
from experimaestro.launchers import Launcher
from experimaestro.scheduler.base import Job

from experimaestro.scheduler.workspace import RunMode
from .context import (
    SerializationContext as SerializationContext,
    SerializedPath as SerializedPath,
    SerializedPathLoader as SerializedPathLoader,
)
from _typeshed import Incomplete
from collections.abc import Generator
from experimaestro.core.types import (
    DeprecatedAttribute as DeprecatedAttribute,
    ObjectType as ObjectType,
)
from experimaestro.utils import logger as logger
from functools import cached_property as cached_property
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import Self

TConfig = TypeVar("TConfig", bound="Config")

class Identifier:
    main: Incomplete
    sub: Incomplete
    def __init__(self, main: bytes, sub: Optional[bytes] = ...) -> None: ...
    def all(self): ...
    def state_dict(self): ...
    @staticmethod
    def from_state_dict(data: Union[Dict[str, str], str]): ...

def is_ignored(value): ...
def remove_meta(value): ...

class ObjectStore: ...

class HashComputer:
    OBJECT_ID: bytes
    INT_ID: bytes
    FLOAT_ID: bytes
    STR_ID: bytes
    PATH_ID: bytes
    NAME_ID: bytes
    NONE_ID: bytes
    LIST_ID: bytes
    TASK_ID: bytes
    DICT_ID: bytes
    ENUM_ID: bytes
    def __init__(self) -> None: ...
    def identifier(self) -> Identifier: ...
    def update(self, value, myself: bool = ...): ...

def updatedependencies(
    dependencies, value: Config, path: List[str], taskids: Set[int]
): ...

class TaggedValue:
    value: Incomplete
    def __init__(self, value) -> None: ...

def add_to_path(p) -> Generator[None, None, None]: ...

class ConfigWalkContext:
    @property
    def path(self) -> None: ...
    def __init__(self) -> None: ...
    @property
    def task(self) -> None: ...
    def currentpath(self) -> Path: ...
    def push(self, key: str): ...

NOT_SET: Incomplete

class ConfigProcessing:
    recurse_task: Incomplete
    visited: Incomplete
    def __init__(self, recurse_task: bool = ...) -> None: ...
    def preprocess(self, config: Config): ...
    def postprocess(self, config: Config, values: Dict[str, Any]): ...
    def list(self, i: int): ...
    def map(self, k: str): ...
    def __call__(self, x): ...

class ConfigWalk(ConfigProcessing):
    context: Incomplete
    def __init__(self, context: ConfigWalkContext) -> None: ...
    def list(self, i: int): ...
    def map(self, k: str): ...

def getqualattr(module, qualname): ...

class ConfigInformation:
    LOADING: bool
    pyobject: Config
    xpmtype: ObjectType
    values: Dict[str, Any]
    job: Job
    dependencies: Incomplete
    def __init__(self, pyobject: TypeConfig) -> None: ...
    def set_meta(self, value: Optional[bool]): ...
    @property
    def meta(self): ...
    def get(self, name): ...
    def set(self, k, v, bypass: bool = ...) -> None: ...
    def addtag(self, name, value) -> None: ...
    def xpmvalues(self, generated: bool = ...) -> Generator[Incomplete, None, None]: ...
    def tags(self): ...
    def validate(self) -> None: ...
    def seal(self, context: ConfigWalkContext): ...
    @property
    def identifier(self) -> Identifier: ...
    def dependency(self): ...
    def updatedependencies(
        self,
        dependencies: Set["experimaestro.dependencies.Dependency"],
        path: List[str],
        taskids: Set[int],
    ): ...
    def submit(
        self,
        workspace,
        launcher,
        run_mode: RunMode = ...,
        init_tasks: List["LightweightTask"] = [],
    ) -> Task: ...
    def outputjson(self, out: io.TextIOBase, context: SerializationContext): ...
    def __json__(self): ...
    def serialize(self, save_directory: Path): ...
    @staticmethod
    def deserialize(
        path: Union[str, Path, SerializedPathLoader], as_instance: bool = ...
    ) -> Config: ...
    @overload
    @staticmethod
    def fromParameters(
        definitions: List[Dict],
        as_instance: bool = ...,
        save_directory: Optional[Path] = ...,
    ) -> TypeConfig: ...
    @overload
    @staticmethod
    def fromParameters(
        definitions: List[Dict],
        as_instance: bool = ...,
        save_directory: Optional[Path] = ...,
    ) -> Config: ...

    class FromPython(ConfigWalk):
        objects: Incomplete
        def __init__(self, context: ConfigWalkContext) -> None: ...
        def preprocess(self, config: Config): ...
        def postprocess(self, config: Config, values: Dict[str, Any]): ...

    def fromConfig(self, context: ConfigWalkContext): ...
    def add_dependencies(self, *dependencies) -> None: ...

def clone(v): ...
def cache(fn, name: str): ...

class TypeConfig:
    __xpmtype__: ObjectType
    __xpm__: Incomplete
    def __init__(self, **kwargs) -> None: ...
    def tag(self, name, value): ...
    def __eq__(self, other): ...
    def __arguments__(self): ...
    def tags(self): ...
    def add_dependencies(self, *dependencies): ...
    def instance(
        self, context: ConfigWalkContext = ..., *, objects: ObjectStore = None
    ) -> Any: ...
    def submit(
        self,
        *,
        workspace: Incomplete | None = ...,
        launcher: Incomplete | None = ...,
        run_mode: RunMode = ...
    ): ...
    def stdout(self): ...
    def stderr(self): ...
    @property
    def job(self): ...
    @property
    def jobpath(self): ...
    def copy(self): ...

@typing_extensions.dataclass_transform(kw_only_default=True)
class Config:
    __xpmtype__: ClassVar[ObjectType]
    __xpm__: ConfigInformation
    __use_xpmobject__: ClassVar[bool]

    XPMValue: Type[Self]
    XPMConfig: Union[Type[Self], Type[TypeConfig[Self]]]
    C: Union[Type[Self], Type[TypeConfig[Self]]]

    @classmethod
    def __getxpmtype__(cls) -> ObjectType: ...
    def __new__(cls, *args, **kwargs) -> Self: ...
    def __validate__(self) -> None: ...
    def __post_init__(self) -> None: ...
    def __json__(self): ...
    def __identifier__(self) -> Identifier: ...
    def add_pretasks(self, *tasks: "LightweightTask"): ...
    def add_pretasks_from(self, configs: "Config"): ...
    def copy_dependencies(self, other: "Config"): ...
    @property
    def pre_tasks(self) -> List["LightweightTask"]: ...

class LightweightTask(Config):
    def execute(self) -> None: ...

class Task(LightweightTask):
    __tags__: Dict[str, str]

    def submit(
        self,
        *,
        workspace: Incomplete | None = ...,
        launcher: Incomplete | None = ...,
        run_mode: RunMode = ...,
        init_tasks: List["LightweightTask"] = []
    ): ...
    def task_outputs(self, dep: Callable[[Config], None]) -> Any: ...

def copyconfig(config_or_output: TConfig, **kwargs) -> TConfig: ...
def setmeta(config: TConfig, flag: bool) -> TConfig: ...

class TypeConfig(Generic[T]):
    def __validate__(self):
        pass



"""Test snippets of code"""

import re
import pytest
from pathlib import Path
import tempfile
import subprocess
import sys
import shlex
import os
import logging

DIR = Path(__file__).parents[3]
RE_SNIPPET_START = re.compile(
    r"<!-- SNIPPET: (?P<id>\S+)"
    r"(?:\s+ARGS\[(?P<args>[^\]]+)\])?(?:\s+ENV\[(?P<env>[^\]]+)\])? -->"
)
RE_SNIPPET_START2 = re.compile(r"```python")
RE_SNIPPET_END = re.compile(r"```")
RE_SNIPPET_VAR = re.compile(r"%([_\w\d]+)%")


class Methods:
    def __iter__(self):
        yield DIR / "README.md", "MAIN"


@pytest.fixture(scope="session")
def snippetpath(tmp_path_factory):
    path = tmp_path_factory.mktemp("snippets")
    return path


def argsub(snippetpath):
    workdir = tempfile.mkdtemp(dir=snippetpath)

    def fn(match):
        varname = match.group(1)
        if varname == "WORKDIR":
            return workdir
        else:
            raise Exception("Unknown variable %s", varname)

    return fn


def get_snippet(snippetpath, path, id):
    with path.open("r") as fp:
        snippet = ""
        args = []
        env = {}
        inside = 0
        for line in fp:
            # logging.info("Line: %s", line.strip())
            m = RE_SNIPPET_START.match(line)
            if m and m.group(1) == id:
                inside = 1
                args = (
                    [
                        RE_SNIPPET_VAR.sub(argsub(snippetpath), arg)
                        for arg in shlex.split(m.group("args"))
                    ]
                    if m.group(2)
                    else []
                )
                env = (
                    {
                        name: val
                        for name, val in (
                            nameval.split("=") for nameval in m.group("env").split(",")
                        )
                    }
                    if m.group("env")
                    else {}
                )
            if inside == 1:
                if RE_SNIPPET_START2.match(line):
                    inside = 2
            elif inside == 2:
                if RE_SNIPPET_END.match(line):
                    inside = False
                else:
                    snippet += line

        return snippet, args, env


@pytest.mark.parametrize("path,id", Methods())
def test_snippet(snippetpath, path, id):
    snippet, args, env = get_snippet(snippetpath, path, id)
    assert snippet != ""

    WAIT_TIME = 20

    fullenv = {name: value for name, value in os.environ.items()}
    fullenv.update(
        {"XPM_ENABLEFAULTHANDLER": "1", "XPM_LOCK_TIMEOUT": str(WAIT_TIME / 2)}
    )
    fullenv.update(env)
    with tempfile.NamedTemporaryFile(
        "wt", suffix=".py", dir=snippetpath, delete=False
    ) as fp:
        fp.write(snippet)

    cmd = [sys.executable, fp.name] + args
    logging.info("Running %s with env=%s", cmd, env)
    p = subprocess.Popen(cmd, env=fullenv)

    try:
        p.wait(WAIT_TIME)
    finally:
        if p.poll() is None:
            logging.error("Killing process PID %s", p.pid)
            p.kill()

    assert p.returncode == 0



from pathlib import Path

import pytest
from experimaestro import Config, Task, Annotated, copyconfig, default
from experimaestro.core.arguments import Param
from experimaestro.core.objects import TypeConfig
from experimaestro.core.types import XPMValue
from experimaestro.generators import pathgenerator
from experimaestro.scheduler.workspace import RunMode
from experimaestro.tests.utils import TemporaryExperiment


@pytest.fixture()
def xp():
    with TemporaryExperiment("deprecated", maxwait=0, run_mode=RunMode.DRY_RUN) as xp:
        yield xp


class A(Config):
    x: Param[int] = 3


def test_object_default():
    """Test plain default value"""
    a = A()
    assert a.x == 3


class B(Config):
    a: Param[A] = A(x=3)


class C(B):
    pass


class D(B, A):
    pass


class DefaultAnnotationConfig(Config):
    a: Annotated[A, default(A(x=3))]


def test_object_config_default():
    """Test default configurations as default values"""
    b = B()
    assert b.a.x == 3

    c = C()
    assert c.a.x == 3

    annotationConfig = DefaultAnnotationConfig()
    assert annotationConfig.a.x == 3


def test_hierarchy():
    """Test if the object hierarchy is OK"""
    OA = A.__getxpmtype__().objecttype
    OB = B.__getxpmtype__().objecttype
    OC = C.__getxpmtype__().objecttype
    OD = D.__getxpmtype__().objecttype

    assert issubclass(A, Config)
    assert issubclass(B, Config)
    assert issubclass(C, Config)

    assert not issubclass(OA, TypeConfig)
    assert not issubclass(OB, TypeConfig)
    assert not issubclass(OC, TypeConfig)

    assert issubclass(C, B)

    assert OA.__bases__ == (A, XPMValue)
    assert OB.__bases__ == (B, XPMValue)
    assert OC.__bases__ == (C, B.XPMValue)
    assert OD.__bases__ == (D, B.XPMValue, A.XPMValue)


class CopyConfig(Task):
    path: Annotated[Path, pathgenerator("hello.txt")]
    x: Param[int]


def test_copyconfig(xp):
    b = CopyConfig(x=2)

    b.submit()

    copy_b = copyconfig(b)

    assert copy_b.x == b.x
    assert "path" not in copy_b.__xpm__.values



# --- Task and types definitions

import logging
from experimaestro import Config, config
from experimaestro.core.objects import TypeConfig

from .utils import TemporaryExperiment
from experimaestro.scheduler import JobState


def test_multiple_inheritance():
    @config()
    class A:
        pass

    @config()
    class B:
        pass

    @config()
    class B1(B):
        pass

    @config()
    class C1(B1, A):
        pass

    @config()
    class C2(A, B1):
        pass

    for C in (C1, C2):
        logging.info("Testing %s", C)
        ctype = C.__xpmtype__
        assert issubclass(C, A)
        assert issubclass(C, B)
        assert issubclass(C, B1)

        assert ctype.objecttype == C.__xpmtype__.objecttype

        assert issubclass(C.__xpmtype__.objecttype, B1.__xpmtype__.basetype)
        assert issubclass(C.__xpmtype__.objecttype, B.__xpmtype__.basetype)
        assert issubclass(C.__xpmtype__.objecttype, A.__xpmtype__.basetype)
        assert not issubclass(C.__xpmtype__.objecttype, TypeConfig)


def test_missing_hierarchy():
    @config()
    class A:
        pass

    class A1(A):
        pass

    @config()
    class B(A1):
        pass

    B.__xpmtype__

    assert issubclass(B, A)
    assert issubclass(B, A1)



from typing import List
from .objects import Task
from .arguments import Param


class TaskSequence(Task):
    """Multiple task execution, the main one being the last"""

    tasks: Param[List[Task]]

    @property
    def __identifier__(self):
        return self.tasks[-1].__identifier__()

    def execute(self):
        self.tasks[-1].__tags__ = self.__tags__
        for task in self.tasks:
            task.execute()


def task_sequence(*tasks: Task):
    return TaskSequence(tasks=tasks)
